/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afh.h structures for audio format handling (para_server) */

/** \cond */
#ifdef HAVE_OGGVORBIS
#define OV_AUDIO_FORMAT " ogg"
#else
#define OV_AUDIO_FORMAT ""
#endif

#ifdef HAVE_FAAD
#define AAC_AUDIO_FORMAT " aac"
#else
#define AAC_AUDIO_FORMAT ""
#endif

#define SUPPORTED_AUDIO_FORMATS "mp3 " OV_AUDIO_FORMAT AAC_AUDIO_FORMAT " wma "

/** \endcond */

/**
 * The tags used by all audio format handlers.
 *
 * Paraslash only uses the more common tags. These are recognized
 * for all supported audio formats.
 */
struct taginfo {
	/** TPE1 (id3v2) / ARTIST (vorbis) / ART (aac) */
	char *artist;
	/** TIT2/TITLE/nam */
	char *title;
	/** TDRC/YEAR/day */
	char *year;
	/** TALB/ALBUM/alb */
	char *album;
	/** COMM/COMMENT/cmt */
	char *comment;
};

/** Audio format dependent information. */
struct afh_info {
	/** The number of chunks this audio file contains. */
	long unsigned chunks_total;
	/** The length of the audio file in seconds. */
	long unsigned seconds_total;
	/** Audio handler specific info about the file. */
	char *techinfo;
	/** Id3 tags, vorbis comments, aac tags. */
	struct taginfo tags;
	/**
  	 * The table that specifies the offset of the individual pieces in
  	 * the current audio file.
 	 */
	uint32_t *chunk_table;
	/** Period of time between sending data chunks. */
	struct timeval chunk_tv;
	/**
	 * The position of the header within the audio file. Ignored if \a
	 * header_len equals zero.
	 */
	uint32_t header_offset;
	/**
	 * The header is needed by senders in case a new client connects in the
	 * middle of the stream. The length of the header defaults to zero
	 * which means that this audio format does not need any special header
	 * treatment. The audio format handler does not need to set this to
	 * zero in this case.
   	 */
	uint32_t header_len;
	/** The number of channels. */
	uint8_t channels;
	/** Frequency in Hz. */
	uint16_t frequency;
	/** Exact meaning depends on audio format. */
	uint16_t bitrate;
};

/**
 *  Structure for audio format handling.
 *
 *  There's one such struct for each supported audio format. Initially, only \a
 *  name and \a init are defined. During the startup process, para_server calls
 *  the \a init function of each audio format handler which is expected to fill
 *  in the other part of this struct.
 */
struct audio_format_handler {
	/** Name of the audio format. */
	const char *name;
	/**
	 * Pointer to the audio format handler's init function.
	 *
	 * Must initialize all function pointers and is assumed to succeed.
	 */
	void (*init)(struct audio_format_handler*);
	/** Typical file endings for files that can be handled by this afh. */
	const char **suffixes;
	/**
	 * Check if this audio format handler can handle the file.
	 *
	 * This is a  pointer to a function returning whether a given file is
	 * valid for this audio format. A negative return value indicates that
	 * this audio format handler is unable to decode the given file. On
	 * success, the function must return a positive value and fill in the
	 * given struct afh_info.
	 *
	 * \sa struct afh_info
	 */
	int (*get_file_info)(char *map, size_t numbytes, int fd,
		struct afh_info *afi);
};

void afh_init(void);
int guess_audio_format(const char *name);
int compute_afhi(const char *path, char *data, size_t size,
	int fd, struct afh_info *afhi);
const char *audio_format_name(int);
void afh_get_chunk(long unsigned chunk_num, struct afh_info *afhi,
		void *map, const char **buf, size_t *len);
uint32_t afh_get_largest_chunk_size(struct afh_info *afhi);
void afh_get_header(struct afh_info *afhi, void *map, const char **buf, size_t *len);
