/*
 * Copyright (C) 1997-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file vss.c The virtual streaming system.
 *
 * This contains the audio streaming code of para_server which is independent
 * of the current audio format, audio file selector and of the activated
 * senders.
 */

#include <regex.h>
#include <dirent.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "portable_io.h"
#include "fec.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "net.h"
#include "server.cmdline.h"
#include "list.h"
#include "vss.h"
#include "send.h"
#include "ipc.h"
#include "fd.h"
#include "sched.h"

extern struct misc_meta_data *mmd;

extern void dccp_send_init(struct sender *);
extern void http_send_init(struct sender *);
extern void udp_send_init(struct sender *);

/** The list of supported senders. */
struct sender senders[] = {
	{
		.name = "http",
		.init = http_send_init,
	},
	{
		.name = "dccp",
		.init = dccp_send_init,
	},
	{
		.name = "udp",
		.init = udp_send_init,
	},
	{
		.name = NULL,
	}
};

/** The possible states of the afs socket. */
enum afs_socket_status {
	/** Socket is inactive. */
	AFS_SOCKET_READY,
	/** Socket fd was included in the write fd set for select(). */
	AFS_SOCKET_CHECK_FOR_WRITE,
	/** vss wrote a request to the socket and waits for reply from afs. */
	AFS_SOCKET_AFD_PENDING
};

/** The task structure for the virtual streaming system. */
struct vss_task {
	/** Copied from the -announce_time command line option. */
	struct timeval announce_tv;
	/** End of the announcing interval. */
	struct timeval data_send_barrier;
	/** End of the EOF interval. */
	struct timeval eof_barrier;
	/** Only used if --autoplay_delay was given. */
	struct timeval autoplay_barrier;
	/** Used for afs-server communication. */
	int afs_socket;
	/** The current state of \a afs_socket. */
	enum afs_socket_status afsss;
	/** The memory mapped audio file. */
	char *map;
	/** Used by the scheduler. */
	struct task task;
	/** Pointer to the header of the mapped audio file. */
	const char *header_buf;
	/** Length of the audio file header. */
	size_t header_len;
	/** Time between audio file headers are sent. */
	struct timeval header_interval;
};

/**
 * The list of currently connected fec clients.
 *
 * Senders may use \ref vss_add_fec_client() to add entries to the list.
 */
static struct list_head fec_client_list;

/**
 * Data associated with one FEC group.
 *
 * A FEC group consists of a fixed number of slices and this number is given by
 * the \a slices_per_group parameter of struct \ref fec_client_parms. Each FEC
 * group contains a number of chunks of the current audio file.
 *
 * FEC slices directly correspond to the data packages sent by the paraslash
 * senders that use FEC. Each slice is identified by its group number and its
 * number within the group. All slices have the same size, but the last slice
 * of the group may not be filled entirely.
 */
struct fec_group {
	/** The number of the FEC group. */
	uint32_t num;
	/** Number of bytes in this group. */
	uint32_t bytes;
	/** The first chunk of the current audio file belonging to the group. */
	uint32_t first_chunk;
	/** The number of chunks contained in this group. */
	uint32_t num_chunks;
	/** When the first chunk was sent. */
	struct timeval start;
	/** The duration of the full group. */
	struct timeval duration;
	/** The group duration divided by the number of slices. */
	struct timeval slice_duration;
	/** Group contains the audio file header that occupies that many slices. */
	uint8_t num_header_slices;
};

/**
 * Describes one connected FEC client.
 */
struct fec_client {
	/** If negative, this client is temporarily disabled. */
	int error;
	/** Parameters requested by the client. */
	struct fec_client_parms *fcp;
	/** Used by the core FEC code. */
	struct fec_parms *parms;
	/** The position of this client in the fec client list. */
	struct list_head node;
	/** When the first slice for this client was sent. */
	struct timeval stream_start;
	/** The first chunk sent to this FEC client. */
	int first_stream_chunk;
	/** Describes the current group. */
	struct fec_group group;
	/** The current slice. */
	uint8_t current_slice_num;
	/** The data to be FEC-encoded (point to a region within the mapped audio file). */
	const unsigned char **src_data;
	/** Last time an audio  header was sent. */
	struct timeval next_header_time;
	/** Used for the last source pointer of an audio file. */
	unsigned char *extra_src_buf;
	/** Extra slices needed to store largest chunk + header. */
	int num_extra_slices;
	/** Contains the FEC-encoded data. */
	unsigned char *enc_buf;
};

/**
 * Get the chunk time of the current audio file.
 *
 * \return A pointer to a struct containing the chunk time, or NULL,
 * if currently no audio file is selected.
 */
struct timeval *vss_chunk_time(void)
{
	if (mmd->afd.afhi.chunk_tv.tv_sec == 0 &&
			mmd->afd.afhi.chunk_tv.tv_usec == 0)
		return NULL;
	return &mmd->afd.afhi.chunk_tv;
}

/**
 * Write a fec header to a buffer.
 *
 * \param buf The buffer to write to.
 * \param h The fec header to write.
 */
static void write_fec_header(struct fec_client *fc, struct vss_task *vsst)
{
	char *buf = (char *)fc->enc_buf;
	struct fec_group *g = &fc->group;
	struct fec_client_parms *p = fc->fcp;

	write_u32(buf, FEC_MAGIC);

	write_u8(buf + 4, p->slices_per_group + fc->num_extra_slices);
	write_u8(buf + 5, p->data_slices_per_group + fc->num_extra_slices);
	write_u32(buf + 6, g->num_header_slices? vsst->header_len : 0);

	write_u32(buf + 10, g->num);
	write_u32(buf + 14, g->bytes);

	write_u8(buf + 18, fc->current_slice_num);
	write_u16(buf + 20, p->max_slice_bytes - FEC_HEADER_SIZE);
	write_u8(buf + 22, g->first_chunk? 0 : 1);
	write_u8(buf + 23, vsst->header_len? 1 : 0);
	memset(buf + 24, 0, 7);
}

static int need_audio_header(struct fec_client *fc, struct vss_task *vsst)
{
	if (!mmd->current_chunk) {
		tv_add(now, &vsst->header_interval, &fc->next_header_time);
		return 0;
	}
	if (!vsst->header_buf)
		return 0;
	if (!vsst->header_len)
		return 0;
	if (fc->group.num && tv_diff(&fc->next_header_time, now, NULL) > 0)
		return 0;
	tv_add(now, &vsst->header_interval, &fc->next_header_time);
	return 1;
}

static int num_slices(long unsigned bytes, struct fec_client *fc, uint8_t *result)
{
	unsigned long m = fc->fcp->max_slice_bytes - FEC_HEADER_SIZE;
	unsigned rv, redundant_slices = fc->fcp->slices_per_group
		- fc->fcp->data_slices_per_group;

	if (!m)
		return -E_BAD_CT;
	rv = (bytes + m - 1) / m;
	if (rv + redundant_slices > 255)
		return -E_BAD_CT;
	*result = rv;
	return 1;
}

/* set group start and group duration */
static void set_group_timing(struct fec_client *fc, struct fec_group *g)
{
	struct timeval *chunk_tv = vss_chunk_time();

	tv_scale(g->num_chunks, chunk_tv, &g->duration);
	tv_divide(fc->fcp->slices_per_group + fc->num_extra_slices,
		&g->duration, &g->slice_duration);
	PARA_DEBUG_LOG("durations (group/chunk/slice): %lu/%lu/%lu\n",
		tv2ms(&g->duration), tv2ms(chunk_tv), tv2ms(&g->slice_duration));
}

static int setup_next_fec_group(struct fec_client *fc, struct vss_task *vsst)
{
	int ret, i, k, data_slices;
	size_t len;
	const char *buf, *start_buf;
	struct fec_group *g = &fc->group;
	unsigned slice_bytes = fc->fcp->max_slice_bytes - FEC_HEADER_SIZE;
	uint32_t max_data_size;

	if (fc->first_stream_chunk < 0) {
		uint8_t hs, ds; /* needed header/data slices */
		uint8_t rs = fc->fcp->slices_per_group
			- fc->fcp->data_slices_per_group; /* redundant slices */
		int n;

		ret = num_slices(vsst->header_len, fc, &hs);
		if (ret < 0)
			return ret;
		ret = num_slices(afh_get_largest_chunk_size(&mmd->afd.afhi),
			fc, &ds);
		if (ret < 0)
			return ret;
		k = (int)hs + ds;
		if (k > 255)
			return -E_BAD_CT;
		if (k < fc->fcp->data_slices_per_group)
			k = fc->fcp->data_slices_per_group;
		n = k + rs;
		fc->num_extra_slices = k - fc->fcp->data_slices_per_group;
		PARA_NOTICE_LOG("fec parms %d:%d:%d (%d extra slices)\n",
			slice_bytes, k, n, fc->num_extra_slices);
		fec_free(fc->parms);
		fc->src_data = para_realloc(fc->src_data, k * sizeof(char *));
		ret = fec_new(k, n, &fc->parms);
		if (ret < 0)
			return ret;
		fc->stream_start = *now;
		fc->first_stream_chunk = mmd->current_chunk;
		g->first_chunk = mmd->current_chunk;
		g->num = 0;
		g->start = *now;
	} else {
		/*
		 * Start and duration of this group depend only on the previous
		 * group. Compute the new group start as g->start += g->duration.
		 */
		struct timeval tmp = g->start;
		tv_add(&tmp, &g->duration, &g->start);
		k = fc->fcp->data_slices_per_group + fc->num_extra_slices;
		set_group_timing(fc, g);
		g->first_chunk += g->num_chunks;
		g->num++;
	}
	if (g->first_chunk >= mmd->afd.afhi.chunks_total)
		return 0;
	if (need_audio_header(fc, vsst)) {
		ret = num_slices(vsst->header_len, fc, &g->num_header_slices);
		if (ret < 0)
			return ret;
	} else
		g->num_header_slices = 0;
	afh_get_chunk(g->first_chunk, &mmd->afd.afhi, vsst->map, &start_buf,
		&len);
	data_slices = k - g->num_header_slices;
	assert(data_slices);
	max_data_size = slice_bytes * data_slices;
	g->bytes = 0;
	for (i = g->first_chunk; i < mmd->afd.afhi.chunks_total; i++) {
		afh_get_chunk(i, &mmd->afd.afhi, vsst->map, &buf, &len);
		if (g->bytes + len > max_data_size)
			break;
		g->bytes += len;
	}
	g->num_chunks = i - g->first_chunk;
	assert(g->num_chunks);
	fc->current_slice_num = 0;
	if (g->num == 0)
		set_group_timing(fc, g);

	/* setup header slices */
	buf = vsst->header_buf;
	for (i = 0; i < g->num_header_slices; i++) {
		fc->src_data[i] = (const unsigned char *)buf;
		buf += slice_bytes;
	}

	/* setup data slices */
	buf = start_buf;
	for (i = g->num_header_slices; i < k; i++) {
		if (buf + slice_bytes > vsst->map + mmd->size)
			/*
			 * Can not use the memory mapped audio file for this
			 * slice as it goes beyond the map. This slice will not
			 * be fully used.
			 */
			break;
		fc->src_data[i] = (const unsigned char *)buf;
		buf += slice_bytes;
	}
	if (i < k) {
		uint32_t payload_size = vsst->map + mmd->size - buf;
		memcpy(fc->extra_src_buf, buf, payload_size);
		fc->src_data[i] = fc->extra_src_buf;
		i++;
		/* use arbitrary data for all remaining slices */
		buf = vsst->map;
		for (; i < k; i++)
			fc->src_data[i] = (const unsigned char *)buf;
	}
	PARA_DEBUG_LOG("FEC group %d: %d chunks (%d - %d), "
		"%d header slices, %d data slices\n",
		g->num, g->num_chunks, g->first_chunk,
		g->first_chunk + g->num_chunks - 1,
		g->num_header_slices, data_slices
	);
	return 1;
}

static int compute_next_fec_slice(struct fec_client *fc, struct vss_task *vsst)
{
	assert(fc->error >= 0);
	if (fc->first_stream_chunk < 0 || fc->current_slice_num
			== fc->fcp->slices_per_group + fc->num_extra_slices) {
		int ret = setup_next_fec_group(fc, vsst);
		if (ret == 0)
			return 0;
		if (ret < 0) {
			PARA_ERROR_LOG("%s\n", para_strerror(-ret));
			PARA_ERROR_LOG("FEC client temporarily disabled\n");
			fc->error = ret;
			return fc->error;
		}
	}
	write_fec_header(fc, vsst);
	fec_encode(fc->parms, fc->src_data, fc->enc_buf + FEC_HEADER_SIZE,
		fc->current_slice_num,
		fc->fcp->max_slice_bytes - FEC_HEADER_SIZE);
	return 1;
}

/**
 * Return a buffer that marks the end of the stream.
 *
 * \param buf Result pointer.
 * \return The length of the eof buffer.
 *
 * This is used for (multicast) udp streaming where closing the socket on the
 * sender might not give rise to an eof condition at the peer.
 */
size_t vss_get_fec_eof_packet(const char **buf)
{
	static const char fec_eof_packet[FEC_HEADER_SIZE] = FEC_EOF_PACKET;
	*buf = fec_eof_packet;
	return FEC_HEADER_SIZE;
}

/**
 * Add one entry to the list of active fec clients.
 *
 * \param fcp Describes the fec parameters to be used for this client.
 * \param result An opaque pointer that must be used by remove the client later.
 *
 * \return Standard.
 */
int vss_add_fec_client(struct fec_client_parms *fcp, struct fec_client **result)
{
	int ret;
	struct fec_client *fc;

	if (fcp->max_slice_bytes < FEC_HEADER_SIZE + fcp->data_slices_per_group)
		return -ERRNO_TO_PARA_ERROR(EINVAL);
	fc = para_calloc(sizeof(*fc));
	fc->fcp = fcp;
	ret = fec_new(fcp->data_slices_per_group, fcp->slices_per_group,
		&fc->parms);
	if (ret < 0)
		goto err;
	fc->first_stream_chunk = -1; /* stream not yet started */
	fc->src_data = para_malloc(fc->fcp->slices_per_group * sizeof(char *));
	fc->enc_buf = para_calloc(fc->fcp->max_slice_bytes);
	fc->num_extra_slices = 0;
	fc->extra_src_buf = para_calloc(fc->fcp->max_slice_bytes);
	fc->next_header_time.tv_sec = 0;
	para_list_add(&fc->node, &fec_client_list);
	*result = fc;
	return 1;
err:
	fec_free(fc->parms);
	free(fc);
	*result = NULL;
	return ret;
}

/**
 * Remove one entry from the list of active fec clients.
 *
 * \param fc The client to be removed.
 */
void vss_del_fec_client(struct fec_client *fc)
{
	list_del(&fc->node);
	free(fc->src_data);
	free(fc->enc_buf);
	free(fc->extra_src_buf);
	fec_free(fc->parms);
	free(fc);
}

/*
 * Compute if/when next slice is due. If it isn't due yet and \a diff is
 * not \p Null, compute the time difference next - now, where
 *
 *	next = stream_start + (first_group_chunk - first_stream_chunk)
 *		* chunk_time + slice_num * slice_time
 */
static int next_slice_is_due(struct fec_client *fc, struct timeval *diff)
{
	struct timeval tmp, next;
	int ret;

	if (fc->first_stream_chunk < 0)
		return 1;
	tv_scale(fc->current_slice_num, &fc->group.slice_duration, &tmp);
	tv_add(&tmp, &fc->group.start, &next);
	ret = tv_diff(&next, now, diff);
	return ret < 0? 1 : 0;
}

static void compute_slice_timeout(struct timeval *timeout)
{
	struct fec_client *fc;

	assert(vss_playing());
	list_for_each_entry(fc, &fec_client_list, node) {
		struct timeval diff;

		if (fc->error < 0)
			continue;
		if (next_slice_is_due(fc, &diff)) {
			timeout->tv_sec = 0;
			timeout->tv_usec = 0;
			return;
		}
		/* timeout = min(timeout, diff) */
		if (tv_diff(&diff, timeout, NULL) < 0)
			*timeout = diff;
	}
}

static void set_eof_barrier(struct vss_task *vsst)
{
	struct fec_client *fc;
	struct timeval timeout = {1, 0}, *chunk_tv = vss_chunk_time();

	if (!chunk_tv)
		goto out;
	list_for_each_entry(fc, &fec_client_list, node) {
		struct timeval group_duration;

		if (fc->error < 0)
			continue;
		tv_scale(fc->group.num_chunks, chunk_tv, &group_duration);
		if (tv_diff(&timeout, &group_duration, NULL) < 0)
			timeout = group_duration;
	}
out:
	tv_add(now, &timeout, &vsst->eof_barrier);
}

/**
 * Check if vss status flag \a P (playing) is set.
 *
 * \return Greater than zero if playing, zero otherwise.
 *
 */
unsigned int vss_playing(void)
{
	return mmd->new_vss_status_flags & VSS_PLAYING;
}

/**
 * Check if the \a N (next) status flag is set.
 *
 * \return Greater than zero if set, zero if not.
 *
 */
unsigned int vss_next(void)
{
	return mmd->new_vss_status_flags & VSS_NEXT;
}

/**
 * Check if a reposition request is pending.
 *
 * \return Greater than zero if true, zero otherwise.
 *
 */
unsigned int vss_repos(void)
{
	return mmd->new_vss_status_flags & VSS_REPOS;
}

/**
 * Check if the vss is currently paused.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_paused(void)
{
	return !(mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

/**
 * Check if the vss is currently stopped.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_stopped(void)
{
	return (mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

static int chk_barrier(const char *bname, const struct timeval *barrier,
		struct timeval *diff, int print_log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (print_log && ms)
		PARA_DEBUG_LOG("%s barrier: %lims left\n", bname, ms);
	return -1;
}

/*
 * != NULL: timeout for next chunk
 * NULL: nothing to do
 */
static struct timeval *vss_compute_timeout(struct vss_task *vsst)
{
	static struct timeval the_timeout;
	struct timeval next_chunk;

	if (vss_next() && vsst->map) {
		/* only sleep a bit, nec*/
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 100;
		return &the_timeout;
	}
	if (chk_barrier("autoplay_delay", &vsst->autoplay_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("eof", &vsst->eof_barrier, &the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("data send", &vsst->data_send_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (!vss_playing() || !vsst->map)
		return NULL;
	compute_chunk_time(mmd->chunks_sent, &mmd->afd.afhi.chunk_tv,
		&mmd->stream_start, &next_chunk);
	if (chk_barrier("chunk", &next_chunk, &the_timeout, 0) >= 0) {
		/* chunk is due or bof */
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 0;
		return &the_timeout;
	}
	/* compute min of current timeout and next slice time */
	compute_slice_timeout(&the_timeout);
	return &the_timeout;
}

static void vss_eof(struct vss_task *vsst)
{

	if (!vsst->map)
		return;
	if (mmd->new_vss_status_flags & VSS_NOMORE)
		mmd->new_vss_status_flags = VSS_NEXT;
	set_eof_barrier(vsst);
	para_munmap(vsst->map, mmd->size);
	vsst->map = NULL;
	mmd->chunks_sent = 0;
	mmd->offset = 0;
	mmd->afd.afhi.seconds_total = 0;
	mmd->afd.afhi.chunk_tv.tv_sec = 0;
	mmd->afd.afhi.chunk_tv.tv_usec = 0;
	free(mmd->afd.afhi.chunk_table);
	mmd->afd.afhi.chunk_table = NULL;
	mmd->mtime = 0;
	mmd->size = 0;
	mmd->events++;
}

/**
 * Get the list of all supported audio formats.
 *
 * \return Aa space separated list of all supported audio formats
 * It is not allocated at runtime, i.e. there is no need to free
 * the returned string in the caller.
 */
const char *supported_audio_formats(void)
{
	return SUPPORTED_AUDIO_FORMATS;
}

static int need_to_request_new_audio_file(struct vss_task *vsst)
{
	struct timeval diff;

	if (vsst->map) /* have audio file */
		return 0;
	if (!vss_playing()) /* don't need one */
		return 0;
	if (mmd->new_vss_status_flags & VSS_NOMORE)
		return 0;
	if (vsst->afsss == AFS_SOCKET_AFD_PENDING) /* already requested one */
		return 0;
	if (chk_barrier("autoplay_delay", &vsst->autoplay_barrier,
			&diff, 1) < 0)
		return 0;
	return 1;
}

/**
 * Compute the timeout for the main select-loop of the scheduler.
 *
 * \param s Pointer to the server scheduler.
 * \param t Pointer to the vss task structure.
 *
 * Before the timeout is computed, the current vss status flags are evaluated
 * and acted upon by calling appropriate functions from the lower layers.
 * Possible actions include
 *
 * 	- request a new audio file from afs,
 *	- shutdown of all senders (stop/pause command),
 *	- reposition the stream (ff/jmp command).
 */
static void vss_pre_select(struct sched *s, struct task *t)
{
	int i;
	struct timeval *tv, diff;
	struct vss_task *vsst = container_of(t, struct vss_task, task);

	if (!vsst->map || vss_next() || vss_paused() || vss_repos()) {
		struct fec_client *fc, *tmp;
		for (i = 0; senders[i].name; i++)
			if (senders[i].shutdown_clients)
				senders[i].shutdown_clients();
		list_for_each_entry_safe(fc, tmp, &fec_client_list, node) {
			fc->first_stream_chunk = -1;
			fc->error = 0;
		}
		mmd->stream_start.tv_sec = 0;
		mmd->stream_start.tv_usec = 0;
	}
	if (vss_next())
		vss_eof(vsst);
	else if (vss_paused()) {
		if (mmd->chunks_sent)
			set_eof_barrier(vsst);
		mmd->chunks_sent = 0;
	} else if (vss_repos()) {
		tv_add(now, &vsst->announce_tv, &vsst->data_send_barrier);
		set_eof_barrier(vsst);
		mmd->chunks_sent = 0;
		mmd->current_chunk = mmd->repos_request;
		mmd->new_vss_status_flags &= ~VSS_REPOS;
	}
	if (need_to_request_new_audio_file(vsst)) {
		PARA_DEBUG_LOG("ready and playing, but no audio file\n");
		para_fd_set(vsst->afs_socket, &s->wfds, &s->max_fileno);
		vsst->afsss = AFS_SOCKET_CHECK_FOR_WRITE;
	} else
		para_fd_set(vsst->afs_socket, &s->rfds, &s->max_fileno);
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].pre_select)
			continue;
		senders[i].pre_select(&s->max_fileno, &s->rfds, &s->wfds);
	}
	tv = vss_compute_timeout(vsst);
	if (tv && tv_diff(tv, &s->timeout, &diff) < 0)
		s->timeout = *tv;
}

static int recv_afs_msg(int afs_socket, int *fd, uint32_t *code, uint32_t *data)
{
	char control[255], buf[8];
	struct msghdr msg = {.msg_iov = NULL};
	struct cmsghdr *cmsg;
	struct iovec iov;
	int ret = 0;

	*fd = -1;
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	memset(buf, 0, sizeof(buf));
	ret = recvmsg(afs_socket, &msg, 0);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (iov.iov_len != sizeof(buf))
		return -E_AFS_SHORT_READ;
	*code = *(uint32_t*)buf;
	*data =  *(uint32_t*)(buf + 4);
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_SOCKET
			|| cmsg->cmsg_type != SCM_RIGHTS)
			continue;
		if ((cmsg->cmsg_len - CMSG_LEN(0)) / sizeof(int) != 1)
			continue;
		*fd = *(int *)CMSG_DATA(cmsg);
	}
	return 1;
}

static void recv_afs_result(struct vss_task *vsst)
{
	int ret, passed_fd, shmid;
	uint32_t afs_code = 0, afs_data = 0;
	struct stat statbuf;

	vsst->afsss = AFS_SOCKET_READY;
	ret = recv_afs_msg(vsst->afs_socket, &passed_fd, &afs_code, &afs_data);
	if (ret < 0)
		goto err;
	PARA_DEBUG_LOG("fd: %d, code: %u, shmid: %u\n", passed_fd, afs_code,
		afs_data);
	ret = -E_NOFD;
	if (afs_code != NEXT_AUDIO_FILE)
		goto err;
	if (passed_fd < 0)
		goto err;
	shmid = afs_data;
	ret = load_afd(shmid, &mmd->afd);
	if (ret < 0)
		goto err;
	shm_destroy(shmid);
	ret = fstat(passed_fd, &statbuf);
	if (ret < 0) {
		PARA_ERROR_LOG("fstat error:\n");
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	mmd->size = statbuf.st_size;
	mmd->mtime = statbuf.st_mtime;
	ret = para_mmap(mmd->size, PROT_READ, MAP_PRIVATE, passed_fd,
		0, &vsst->map);
	if (ret < 0)
		goto err;
	close(passed_fd);
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	mmd->events++;
	mmd->num_played++;
	mmd->new_vss_status_flags &= (~VSS_NEXT);
	afh_get_header(&mmd->afd.afhi, vsst->map, &vsst->header_buf,
		&vsst->header_len);
	return;
err:
	free(mmd->afd.afhi.chunk_table);
	if (passed_fd >= 0)
		close(passed_fd);
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	mmd->new_vss_status_flags = VSS_NEXT;
}

/**
 * Main sending function.
 *
 * This function gets called from vss_post_select(). It checks whether the next
 * chunk of data should be pushed out. It obtains a pointer to the data to be
 * sent out as well as its length from mmd->afd.afhi. This information is then
 * passed to each supported sender's send() function as well as to the send()
 * functions of each registered fec client.
 */
static void vss_send(struct vss_task *vsst)
{
	int i, fec_active = 0;
	struct timeval due;
	struct fec_client *fc, *tmp_fc;

	if (!vsst->map || !vss_playing())
		return;
	if (chk_barrier("eof", &vsst->eof_barrier, &due, 1) < 0)
		return;
	if (chk_barrier("data send", &vsst->data_send_barrier,
			&due, 1) < 0)
		return;
	list_for_each_entry_safe(fc, tmp_fc, &fec_client_list, node) {
		if (fc->error < 0)
			continue;
		if (!next_slice_is_due(fc, NULL)) {
			fec_active = 1;
			continue;
		}
		if (compute_next_fec_slice(fc, vsst) <= 0)
			continue;
		PARA_DEBUG_LOG("sending %d:%d (%u bytes)\n", fc->group.num,
			fc->current_slice_num, fc->fcp->max_slice_bytes);
		fc->fcp->send((char *)fc->enc_buf,
			fc->fcp->max_slice_bytes,
			fc->fcp->private_data);
		fc->current_slice_num++;
		fec_active = 1;
	}
	if (mmd->current_chunk >= mmd->afd.afhi.chunks_total) { /* eof */
		if (!fec_active)
			mmd->new_vss_status_flags |= VSS_NEXT;
		return;
	}
	compute_chunk_time(mmd->chunks_sent, &mmd->afd.afhi.chunk_tv,
		&mmd->stream_start, &due);
	if (tv_diff(&due, now, NULL) <= 0) {
		const char *buf;
		size_t len;

		if (!mmd->chunks_sent) {
			struct timeval tmp;
			mmd->stream_start = *now;
			tv_scale(mmd->current_chunk, &mmd->afd.afhi.chunk_tv, &tmp);
			mmd->offset = tv2ms(&tmp);
			mmd->events++;
		}
		/*
		 * We call the send function also in case of empty chunks as
		 * they might have still some data queued which can be sent in
		 * this case.
		 */
		afh_get_chunk(mmd->current_chunk, &mmd->afd.afhi, vsst->map,
			&buf, &len);
		for (i = 0; senders[i].name; i++) {
			if (!senders[i].send)
				continue;
			senders[i].send(mmd->current_chunk, mmd->chunks_sent,
				buf, len, vsst->header_buf, vsst->header_len);
		}
		mmd->chunks_sent++;
		mmd->current_chunk++;
	}
}

static void vss_post_select(struct sched *s, struct task *t)
{
	int ret, i;
	struct vss_task *vsst = container_of(t, struct vss_task, task);


	if (mmd->sender_cmd_data.cmd_num >= 0) {
		int num = mmd->sender_cmd_data.cmd_num,
			sender_num = mmd->sender_cmd_data.sender_num;

		if (senders[sender_num].client_cmds[num])
			senders[sender_num].client_cmds[num](&mmd->sender_cmd_data);
		mmd->sender_cmd_data.cmd_num = -1;
	}
	if (vsst->afsss != AFS_SOCKET_CHECK_FOR_WRITE) {
		if (FD_ISSET(vsst->afs_socket, &s->rfds))
			recv_afs_result(vsst);
	} else if (FD_ISSET(vsst->afs_socket, &s->wfds)) {
		PARA_NOTICE_LOG("requesting new fd from afs\n");
		ret = send_buffer(vsst->afs_socket, "new");
		if (ret < 0)
			PARA_CRIT_LOG("%s\n", para_strerror(-ret));
		else
			vsst->afsss = AFS_SOCKET_AFD_PENDING;
	}
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].post_select)
			continue;
		senders[i].post_select(&s->rfds, &s->wfds);
	}
	if ((vss_playing() && !(mmd->vss_status_flags & VSS_PLAYING)) ||
			(vss_next() && vss_playing()))
		tv_add(now, &vsst->announce_tv, &vsst->data_send_barrier);
	vss_send(vsst);
}

/**
 * Initialize the virtual streaming system task.
 *
 * \param afs_socket The fd for communication with afs.
 *
 * This also initializes all supported senders and starts streaming
 * if the --autoplay command line flag was given.
 */
void init_vss_task(int afs_socket)
{
	static struct vss_task vss_task_struct, *vsst = &vss_task_struct;
	int i;
	char *hn = para_hostname(), *home = para_homedir();
	long unsigned announce_time = conf.announce_time_arg > 0?
			conf.announce_time_arg : 300,
		autoplay_delay = conf.autoplay_delay_arg > 0?
			conf.autoplay_delay_arg : 0;
	vsst->header_interval.tv_sec = 5; /* should this be configurable? */
	vsst->afs_socket = afs_socket;
	vsst->task.pre_select = vss_pre_select;
	vsst->task.post_select = vss_post_select;
	ms2tv(announce_time, &vsst->announce_tv);
	PARA_INFO_LOG("announce timeval: %lums\n", tv2ms(&vsst->announce_tv));
	INIT_LIST_HEAD(&fec_client_list);
	for (i = 0; senders[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s sender\n", senders[i].name);
		senders[i].init(&senders[i]);
	}
	free(hn);
	free(home);
	mmd->sender_cmd_data.cmd_num = -1;
	if (conf.autoplay_given) {
		struct timeval tmp;
		mmd->vss_status_flags |= VSS_PLAYING;
		mmd->new_vss_status_flags |= VSS_PLAYING;
		ms2tv(autoplay_delay, &tmp);
		tv_add(now, &tmp, &vsst->autoplay_barrier);
		tv_add(&vsst->autoplay_barrier, &vsst->announce_tv,
			&vsst->data_send_barrier);
	}
	register_task(&vsst->task);
}
